import hopsworks

from typing import Annotated
from langchain_openai import ChatOpenAI
from typing_extensions import TypedDict

from langchain_core.tools import tool
from langchain_core.messages import ToolMessage
from langgraph.graph import StateGraph, START, END
from langgraph.graph.message import add_messages
from langchain_core.runnables.config import RunnableConfig


class State(TypedDict):
    # Messages have the type "list". The `add_messages` function
    # in the annotation defines how this state key should be updated
    # (in this case, it appends messages to the list, rather than overwriting them)
    messages: Annotated[list, add_messages]


class ChatbotAgent:
    def __init__(self):
        self._tools = [ChatbotAgent.retrieve_change_request]
        self._tools_by_name = {tool.name: tool for tool in self._tools}

        project = hopsworks.login(
            host="57.129.57.102",
            port=28181,
            project="ericsson",
            api_key_value="WetaZiKuXKBRxczc.WVITTNj71JV57oYcHaOVGOC5WHXra2i2tOaJ6d68fNCwAhBP2i5qUqq9YLZucn54"
        )

        self._feature_store = project.get_feature_store()
        self._setup_graph()

    @tool
    def retrieve_change_request(change_request: str, config: RunnableConfig) -> str:
        """Retrieve information about a change request
        
        # Arguments: 
            change_request: the chane request to retrieve 
        
        # Return the metadata object for the change request
        """
        print("the tool was called")
        return ToolMessage(
            content="These are the very important meeting notes",
            name="something",
            tool_call_id="somethingelse",
        )

    def _parse_tools_call(self, message):
        # Remove Qwen special tool chars
        import json

        message = message.replace("<tool_call>", "").replace("</tool_call>", "")
        return [json.loads(message)]

    def _tool_node(self, state: dict):
        """Method to invoke the tool identified by the LLM"""
        print("calling tool")
        result = []
        for tool_call in self._parse_tools_call(state["messages"][-1].content):
            tool = self._tools_by_name[tool_call["name"]]
            result.append(tool.invoke(tool_call["arguments"]))
        return {"messages": result}

    def _route_tools(
        self,
        state: State,
    ):
        """
        Use in the conditional_edge to route to the ToolNode if the last message
        has tool calls. Otherwise, route to the end.
        """
        if isinstance(state, list):
            ai_message = state[-1]
        elif messages := state.get("messages", []):
            ai_message = messages[-1]
        else:
            raise ValueError(f"No messages found in input state to tool_edge: {state}")
        if "<tool_call>" in ai_message.content:
            print("tool is supposed to be invoked")
            return "tools"
        return END

    def _chatbot(self, state: State):
        print("this is the chatbot")
        return {"messages": [self._llm.invoke(state["messages"])]}

    def _setup_graph(self):
        self._llm = ChatOpenAI(
            base_url="http://57.129.56.184/openai/v1",
            model_name="qwen257binstruct",
            temperature=0,
            max_tokens=1024,
            top_p=1.0,
            api_key="X",
            default_headers={
                "Authorization": "ApiKey WetaZiKuXKBRxczc.WVITTNj71JV57oYcHaOVGOC5WHXra2i2tOaJ6d68fNCwAhBP2i5qUqq9YLZucn54",
                "Host": "qwen257binstruct.ericsson.hopsworks.ai",
            },
        ).bind_tools(self._tools)

        graph_builder = StateGraph(State)

        graph_builder.add_node("chatbot", self._chatbot)
        graph_builder.add_node("tools", self._tool_node)

        graph_builder.add_conditional_edges(
            "chatbot",
            self._route_tools,
            # The following dictionary lets you tell the graph to interpret the condition's outputs as a specific node
            # It defaults to the identity function, but if you
            # want to use a node named something else apart from "tools",
            # You can update the value of the dictionary to something else
            # e.g., "tools": "my_tools"
            {"tools": "tools", END: END},
        )
        graph_builder.add_edge("tools", "chatbot")
        graph_builder.add_edge(START, "chatbot")

        self._graph = graph_builder.compile()
    
    def invoke(self, user_query):

        """Invoke the Chatbot"""

        system_message = """
        You are a chatbot designed to manage the meeting notes for the 5G committee meetings.
        You should use the tools provided to retrieve the relevant notes based on the user question, summarize them and return the information to the user
        """
        
        answers = self._graph.invoke(
            {"messages": [("system", system_message), ("user", user_query)]}
        )

        return answers["messages"][-1].content
